// This example is from the book Developing Java Beans by Robert Englander. 
// Copyright (c) 1997 O'Reilly & Associates.
// You may study, use, modify, and distribute this example for any purpose.
// This example is provided WITHOUT WARRANTY either expressed or implied.

// Chapter 8 -- The Thermostat class

package BeansBook.Simulator;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;
import java.io.*;

// the Thermostat bean class definition
public class Thermostat extends Panel  // subclasses Panel
    implements PropertyChangeListener, // property changes from temperature source
               ActionListener,         // action events from buttons
               ItemListener            // item events choice controls
{
   // the choice control for display units
   protected Choice unitsChoice;
   
   // the choice control for displaying ambient or comfort level temperature
   protected Choice displayChoice;
   
   // button for decrementing comfort level temperature
   protected Button decButton;
   
   // button for incrementing comfort level temperature
   protected Button incButton;
   
   // label for displaying temperatures
   protected Label  displayLabel;

   // flags for the NeedsCooling and NeedsHeating state
   protected boolean bNeedsCooling = false;
   protected boolean bNeedsHeating = false;
   
   // the comfort level temperature
   protected double comfortTemp = 22.0;
   
   // the ambient temperature from the temperature source
   protected double ambientTemp;
   
   // flag indicating if ambient temperature is being displayed
   protected boolean bShowingAmbient = true;
   
   // flag indicating if Celsius units are being displayed
   protected boolean bShowingCelsius = true;

   // support for the constrained properties
   protected VetoableChangeSupport vetoSupport;

   // cooling and heating request listeners
   protected transient Vector coolingListeners;
   protected transient Vector heatingListeners;

   // constructor
   public Thermostat()
   {
      // call superclass constructor
      super();
      
      // create the support object for constrained properties
      vetoSupport = new VetoableChangeSupport(this);
      
      // use a border layout for the constituents
      setLayout(new BorderLayout());
      
      // create the constituents
      unitsChoice = new Choice();
      displayChoice = new Choice();
      decButton = new Button("<<");
      incButton = new Button(">>");
      displayLabel = new Label("**********");

      // register as action listener for the buttons
      decButton.addActionListener(this);
      incButton.addActionListener(this);
      
      // register as item listener for the choice controls
      displayChoice.addItemListener(this);
      unitsChoice.addItemListener(this);
      
      // disable the comfort temperature buttons
      decButton.setEnabled(false);
      incButton.setEnabled(false);

      // add the constituents in their appropriate locations
      add(unitsChoice, BorderLayout.NORTH);
      add(displayChoice, BorderLayout.SOUTH);
      add(decButton, BorderLayout.WEST);
      add(incButton, BorderLayout.EAST);
      add(displayLabel, BorderLayout.CENTER);
      
      // use center alignment for the temperature display
      displayLabel.setAlignment(Label.CENTER);
      
      // add the units to the choice control
      unitsChoice.add("Celsius");
      unitsChoice.add("Fahrenheit");
      
      // add the display types to the choice control
      displayChoice.add("Ambient");
      displayChoice.add("Comfort");
   }

   // handle an action event
   public void actionPerformed(ActionEvent evt)
   {
      // change in temperature
      double tempDelta;
      
      // delta value for rounding
      double roundingDelta;
      
      // it was the decrement button
      if (evt.getSource() == decButton)
      {
         // reduce temp by 1
         tempDelta = -1.0;
         
         // delta is 0 for rounding down
         roundingDelta = 0.0;
      }
      else // it was the increment button
      {
         // increase temp by 1
         tempDelta = 1.0;
         
         // delta is 1 for rounding up
         roundingDelta = 1.0;
      }

      // the new proposed comfort temperature
      double newValue;
      
      // displaying temperatures in Celsius
      if (bShowingCelsius)
      {
         // just add the delta
         newValue = comfortTemp + tempDelta;
      }
      else // displaying in Fahrenheit
      {
         // convert to Fahrenheit, add the delta, and convert back
         double t = 32.0 + ((comfortTemp * 9.0) / 5.0);
         t += tempDelta;
         newValue = (t - 32.0) * 5.0 / 9.0; 
      }
      
      // the old value object for firing the vetoable change event
      Double old = new Double(comfortTemp);
      try
      {
         // fire the event
         vetoSupport.fireVetoableChange("ComfortTemperature", old, new  
                                 Double(newValue));
         
         // if we get this far we can make the change
         synchronized (this)
         {
            comfortTemp = newValue;
         }
      }
      catch (PropertyVetoException e)
      {
         // the change was vetoed by a listening object, but
         // if we have a fractional part, we could try rounding the value
         // to use a whole number
         double wholePart = (double)old.longValue();
         if ((old.doubleValue() - wholePart) == 0.0)
         {
            // we can't make a change
            return;
         }
             
         // attempt to alter the comfort temperature using the whole part of the
         // old value and the rounding delta
         newValue = wholePart + roundingDelta;
         
         try
         {
            // fire the event
            vetoSupport.fireVetoableChange("ComfortTemperature", old, 
                                 new Double(newValue));
         }
         catch (PropertyVetoException ee)
         {
            // we couldnt make this change either
            return;
         }

         // we can go ahead and change it now
         synchronized (this)
         {
            comfortTemp = wholePart + roundingDelta;
         }
      }

      // set the needs cooling and needs heating states, comparing 
      // the ambient and comfort level temperatures
      setNeedsCooling((ambientTemp > comfortTemp) && 
                      ((ambientTemp - comfortTemp) >= 1.0));
      setNeedsHeating((ambientTemp < comfortTemp) && 
                      ((comfortTemp - ambientTemp) >= 1.0));

      // redisplay the temperature
      displayTemp();
   }
   
   // display the temperature
   protected void displayTemp()
   {
      // temporary temperature value
      double t;
      if (bShowingAmbient)
      {
         // use the ambient temperature
         t = ambientTemp;
      }
      else
      {
         // use the comfort level temperature
         t = comfortTemp;
      }
      
      // if not using Celsius, convert to Fahrenheit
      if (!bShowingCelsius)
      {
         t = 32.0 + ((t * 9.0) / 5.0);
      }
      
      // display whole number part of the temperature
      Double d = new Double(t);
      displayLabel.setText(new Long(d.longValue()).toString());
   }

   // handle item state change events
   public void itemStateChanged(ItemEvent evt)
   {
      if (evt.getSource() == displayChoice &&
           evt.getStateChange() == ItemEvent.SELECTED)
      {
         // determine the newly selected item string
         String sel = (String)evt.getItem();
         if (sel.equals("Comfort"))
         {
            // showing comfort level, not ambient
            bShowingAmbient = false;
            
            // enable comfort level buttons
            decButton.setEnabled(true);
            incButton.setEnabled(true);
            
            // display the temperature
            displayTemp();
         }
         else if (sel.equals("Ambient"))
         {
            // showing ambient temperature
            bShowingAmbient = true;
            
            // disable the comfort level buttons
            decButton.setEnabled(false);
            incButton.setEnabled(false);
            
            // display the temperature
            displayTemp();
         }
      }

      else if (evt.getSource() == unitsChoice &&
           evt.getStateChange() == ItemEvent.SELECTED)
      {
         // a new temperature units item has been selected
         Choice pp = (Choice)evt.getSource();
         
         // determine the newly selected item string
         String sel = (String)evt.getItem();
         if (sel.equals("Celsius"))
         {
            // showing Celsius
            bShowingCelsius = true;
            
            // display the temperature
            displayTemp();
         }
         else if (sel.equals("Fahrenheit"))
         {
            // showing Fahrenheit, not Celsius
            bShowingCelsius = false;
            
            // display the temperature
            displayTemp();
         }
      }
   }

   // add a heating request listener
   public synchronized void
   addHeatingRequestListener(HeatingRequestListener l)
   {
      // add a listener if it is not already registered
      if (heatingListeners == null)
      {
         heatingListeners = new Vector();
      }
      
      if (!heatingListeners.contains(l))
      {
         heatingListeners.addElement(l);
      }
   }

   // remove the heating request listener
   public synchronized void 
   removeHeatingRequestListener(HeatingRequestListener l)
   {
      // remove it if it is registered
      if (heatingListeners != null)
      {
         heatingListeners.removeElement(l);
      }  
   }

   // add a cooling request listener
   public synchronized void
   addCoolingRequestListener(CoolingRequestListener l)
   {
      // add a listener if it is not already registered
      if (coolingListeners == null)
      {
         coolingListeners = new Vector();
      }

      if (!coolingListeners.contains(l))
      {
         coolingListeners.addElement(l);
      }
   }

   // remove the cooling request listener
   public synchronized void 
   removeCoolingRequestListener(CoolingRequestListener l)
   {
      // remove it if it is registered
      if (coolingListeners != null)
      {
         coolingListeners.removeElement(l);
      }  
   }

   // return the comfort temperature
   public synchronized double getComfortTemperature()
   {
      return comfortTemp;
   }

   // set the needs cooling state
   private void setNeedsCooling(boolean b)
   {
      // do nothing if there is no change
      if (b == bNeedsCooling)
      {
         return;
      }

      bNeedsCooling = b;

      // nothing else to do if collection isnt allocated yet      
      if (coolingListeners == null)
      {
         return;
      }
      
      // fire the cooling service request
      ServiceRequestEvent evt = new ServiceRequestEvent(this, bNeedsCooling);
      
      // make a copy of the listener object vector so that it cannot
      // be changed while we are firing events
      Vector v;
      synchronized(this)
      {
         v = (Vector) coolingListeners.clone();
      }

      // fire the event to all listeners
      int cnt = v.size();
      for (int i = 0; i < cnt; i++)
      {
         CoolingRequestListener client = (CoolingRequestListener)v.elementAt(i);
         client.coolingRequest(evt);
      }
   }
   
   // set the needs heating state
   private void setNeedsHeating(boolean b)
   {
      // do nothing if there is no change
      if (b == bNeedsHeating)
      {
         return;
      }

      bNeedsHeating = b;

      // nothing else to do if collection isnt allocated yet
      if (heatingListeners == null)
      {
         return;
      }
            
      // fire the heating service request
      ServiceRequestEvent evt = new ServiceRequestEvent(this, bNeedsHeating);
      
      // make a copy of the listener object vector so that it cannot
      // be changed while we are firing events
      Vector v;
      synchronized(this)
      {
         v = (Vector) heatingListeners.clone();
      }

      // fire the event to all listeners
      int cnt = v.size();
      for (int i = 0; i < cnt; i++)
      {
         HeatingRequestListener client = (HeatingRequestListener)v.elementAt(i);
         client.heatingRequest(evt);
      }
   }

   // add vetoable change listener
   public void addVetoableChangeListener(VetoableChangeListener l)
   {
      vetoSupport.addVetoableChangeListener(l);
   }
   
   // remove vetoable change listener
   public void removeVetoableChangeListener(VetoableChangeListener l)
   {
      vetoSupport.removeVetoableChangeListener(l);
   }

   // handle a property change from another source
   public void propertyChange(PropertyChangeEvent evt)
   {
      // check for a Temperature property change
      if (evt.getPropertyName().equals("Temperature"))
      {
         // modify the ambient temperature
         Double d = (Double)evt.getNewValue();
         ambientTemp = d.doubleValue();
         
         // update the bNeedsCooling and bNeedsHeating variables
         setNeedsCooling((ambientTemp > comfortTemp) && 
                        ((ambientTemp - comfortTemp) >= 1.0));
         setNeedsHeating((ambientTemp < comfortTemp) && 
                        ((comfortTemp - ambientTemp) >= 1.0));
         
         // display the temperature
         displayTemp();
      }
   }

   private void writeObject(ObjectOutputStream stream)
               throws java.io.IOException
   {
      // perform default writing first
      stream.defaultWriteObject();

      // if we have allocated coolingListeners
      if (coolingListeners != null)
      {
         // clone the vector in case one is added or removed
         Vector v = null;
         synchronized (this) 
         {
            v = (Vector) coolingListeners.clone();
         }

         int cnt = v.size();
         for(int i = 0; i < cnt; i++) 
         {
            // get the listener element from the collection
            CoolingRequestListener l = 
            (CoolingRequestListener)v.elementAt(i);
            
            // if the listener is serializable, write it to the stream
            if (l instanceof Serializable) 
            {
               stream.writeObject(l);
            }
         }
      }
      
      // a null object marks the end of the cooling request listeners 
      stream.writeObject(null);

      // if we have allocated heatingListeners
      if (heatingListeners != null)
      {
         // clone the vector in case one is added or removed
         Vector v = null;
         synchronized (this) 
         {
            v = (Vector) heatingListeners.clone();
         }

         int cnt = v.size();
         for(int i = 0; i < cnt; i++) 
         {
            // get the listener element from the collection
            HeatingRequestListener l = 
            (HeatingRequestListener)v.elementAt(i);
            
            // if the listener is serializable, write it to the stream
            if (l instanceof Serializable) 
            {
               stream.writeObject(l);
            }
         }
      }
      
      // a null object marks the end of the heating request listeners 
      stream.writeObject(null);
   }

   // handle the reading of the object state
   private void readObject(ObjectInputStream stream)
      throws java.io.IOException
   {
      try
      {
         stream.defaultReadObject();
         
         Object l;

         // get the cooling request listeners
         while(null != (l = stream.readObject())) 
         {
            addCoolingRequestListener((CoolingRequestListener)l);
         }

         // get the heating request listeners
         while(null != (l = stream.readObject())) 
         {
            addHeatingRequestListener((HeatingRequestListener)l);
         }
      }
      catch (ClassNotFoundException e)
      {
         throw new IOException();
      }
   }
}
